@@ -1,4 +1,4 @@
-#! /usr/bin/env python
+#!/usr/bin/env python
 # encoding: utf-8
 # Thomas Nagy, 2006-2008 (ita)
 
@@ -8,13 +8,15 @@
 import re, sys, os, string, types
 if __name__ == '__main__':
 	sys.path = ['.', '..'] + sys.path
-import Params
-from Params import debug, error, warning
+import Logs, Build, Utils
+from Logs import debug, error
 import traceback
 
-class PreprocError(Exception):
+class PreprocError(Utils.WafError):
 	pass
 
+POPFILE = '-'
+
 go_absolute = 0
 "set to 1 to track headers on files in /usr/include - else absolute paths are ignored"
 
@@ -22,9 +24,6 @@ standard_includes = ['/usr/include']
 if sys.platform == "win32":
 	standard_includes = []
 
-g_findall = 1
-'search harder for project includes'
-
 use_trigraphs = 0
 'apply the trigraph rules first'
 
@@ -49,6 +48,17 @@ g_optrans = {
 re_lines = re.compile(\
 	'^[ \t]*(#|%:)[ \t]*(ifdef|ifndef|if|else|elif|endif|include|import|define|undef|pragma)[ \t]*(.*)\r*$',
 	re.IGNORECASE | re.MULTILINE)
+
+# Reasons for using the Waf preprocessor by default
+# 1. the preprocessing is performed once for the clean build, and each time a source file changes (fast)
+# 2. unnecessary rebuilds might occur (#include in comments)
+# 3. some includes use the preprocessor, for example #include A()
+# 4. include guards might not be taken into account, resulting in infinite loops
+# 5. the bugs in the waf preprocessor are usually fixed quickly
+#
+# if you think your project does not need it, use this regexp to catch all includes
+#re_lines = re.compile('^[ \t]*(#|%:)[ \t]*(include|import)[ \t]*(.*)\r*$', re.IGNORECASE | re.MULTILINE)
+
 re_mac = re.compile("^[a-zA-Z_]\w*")
 re_fun = re.compile('^[a-zA-Z_][a-zA-Z0-9_]*[(]')
 re_pragma_once = re.compile('^\s*once\s*', re.IGNORECASE)
@@ -159,10 +169,10 @@ def get_expr(lst, defs, ban):
 			(p2, v2, lst2) = get_expr(lst[1:], defs, ban)
 
 			if v == '#':
-				if p2 != IDENT: raise PreprocError, "ident expected %s" % str(lst)
+				if p2 != IDENT: raise PreprocError("ident expected %s" % str(lst))
 				return get_expr([(STR, v2)]+lst2, defs, ban)
 
-			if p2 != NUM: raise PreprocError, "num expected %s" % str(lst)
+			if p2 != NUM: raise PreprocError("num expected %s" % str(lst))
 
 			if   v == '+': return (p2, v2, lst2)
 			elif v == '-': return (p2, - int(v2), lst2)
@@ -181,7 +191,7 @@ def get_expr(lst, defs, ban):
 				elif v == '(': count_par += 1
 				i += 1
 			else:
-				raise PreprocError, "rparen expected %s" % str(lst)
+				raise PreprocError("rparen expected %s" % str(lst))
 
 			ret = process_tokens(lst[1:i], defs, ban)
 			if len(ret) == 1:
@@ -189,7 +199,7 @@ def get_expr(lst, defs, ban):
 				return (p, v, lst[i+1:])
 			else:
 				#return (None, lst1, lst[i+1:])
-				raise PreprocError, "cannot reduce %s" % str(lst)
+				raise PreprocError("cannot reduce %s" % str(lst))
 
 	elif p == IDENT:
 		if len(lst)>1:
@@ -198,20 +208,20 @@ def get_expr(lst, defs, ban):
 				# token pasting, reevaluate the identifier obtained
 				(p3, v3) = lst[2]
 				if p3 != IDENT and p3 != NUM and p3 != OP:
-					raise PreprocError, "%s: ident expected after '##'" % str(lst)
+					raise PreprocError("%s: ident expected after '##'" % str(lst))
 				return get_expr([(p, v+v3)]+lst[3:], defs, ban)
 
 		if v.lower() == 'defined':
 			(p2, v2) = lst[1]
 			off = 2
 			if v2 == '(':
-				(p3, v3) = lst[2]
-				if p3 != IDENT: raise PreprocError, 'expected an identifier after a "defined("'
-				(p2, v2) = lst[3]
-				if v2 != ')': raise PreprocError, 'expected a ")" after a "defined(x"'
+				(p2, v2) = lst[2]
+				if p2 != IDENT: raise PreprocError('expected an identifier after a "defined("')
+				(p3, v3) = lst[3]
+				if v3 != ')': raise PreprocError('expected a ")" after a "defined(x"')
 				off = 4
 			elif p2 != IDENT:
-				raise PreprocError, 'expected a "(" or an identifier after a defined'
+				raise PreprocError('expected a "(" or an identifier after a defined')
 
 			x = 0
 			if v2 in defs: x = 1
@@ -229,15 +239,15 @@ def get_expr(lst, defs, ban):
 		macro_def = defs[v]
 
 		if not macro_def[0]:
-			# simple macro, substitute, and reevaluate
-			lst = macro_def[1] + lst[1:]
+			# simple macro, substitute, and reevaluate (3: is for removing the name and the two parens)
+			lst = macro_def[1] + lst[3:]
 			return get_expr(lst, defs, ban)
 		else:
 			# collect the arguments for the funcall
 			params = []
 			i = 1
 			p2, v2 = lst[i]
-			if p2 != OP or v2 != '(': raise PreprocError, "invalid function call '%s'" % v
+			if p2 != OP or v2 != '(': raise PreprocError("invalid function call '%s'" % v)
 
 			one_param = []
 			count_paren = 0
@@ -255,7 +265,7 @@ def get_expr(lst, defs, ban):
 							lst = lst[i+1:]
 							break
 						elif v2 == ',':
-							if not one_param: raise PreprocError, "empty param in funcall %s" % p
+							if not one_param: raise PreprocError("empty param in funcall %s" % p)
 							params.append(one_param)
 							one_param = []
 						else:
@@ -266,7 +276,7 @@ def get_expr(lst, defs, ban):
 						elif v2 == ')': count_paren -= 1
 
 			except IndexError, e:
-				#raise PreprocError, 'invalid function call %s: missing ")"' % p
+				#raise PreprocError('invalid function call %s: missing ")"' % p)
 				raise
 
 			# substitute the arguments within the define expression
@@ -308,7 +318,7 @@ def process_tokens(lst, defs, ban):
 
 			op1, ov1 = nlst[0]
 			if op1 != OP:
-				raise PreprocError, "op expected %s" % str(lst)
+				raise PreprocError("op expected %s" % str(lst))
 
 			if ov1 == '?':
 				i = 0
@@ -318,7 +328,7 @@ def process_tokens(lst, defs, ban):
 					elif k == '(': count_par += 1
 					elif k == ':' and count_par == 0: break
 					i += 1
-				else: raise PreprocError, "ending ':' expected %s" % str(lst)
+				else: raise PreprocError("ending ':' expected %s" % str(lst))
 
 				if reduce_nums(v, 0, '+'): lst = nlst[1:i]
 				else: lst = nlst[i+1:]
@@ -329,7 +339,7 @@ def process_tokens(lst, defs, ban):
 				continue
 
 			p2, v2, nlst = get_expr(nlst[1:], defs, ban)
-			if p2 != NUM: raise PreprocError, "num expected after op %s" % str(lst)
+			if p2 != NUM: raise PreprocError("num expected after op %s" % str(lst))
 			if nlst:
 				# op precedence
 				op3, ov3 = nlst[0]
@@ -346,7 +356,7 @@ def process_tokens(lst, defs, ban):
 			continue
 
 		elif p == STR:
-			if nlst: raise PreprocError, "sequence must terminate with a string %s" % str(nlst)
+			if nlst: raise PreprocError("sequence must terminate with a string %s" % str(nlst))
 			return [(p, v)]
 
 		return (None, None, [])
@@ -354,12 +364,12 @@ def process_tokens(lst, defs, ban):
 def eval_macro(lst, adefs):
 	# look at the result, and try to return a 0/1 result
 	ret = process_tokens(lst, adefs, [])
-	if not ret: raise PreprocError, "missing tokens to evaluate %s" % str(lst)
+	if not ret: raise PreprocError("missing tokens to evaluate %s" % str(lst))
 	p, v = ret[0]
 	return int(v) != 0
 
 class c_parser(object):
-	def __init__(self, nodepaths=None, strpaths=None, defines=None):
+	def __init__(self, nodepaths=None, defines=None):
 		#self.lines = txt.split('\n')
 		self.lines = []
 
@@ -371,22 +381,14 @@ class c_parser(object):
 
 		self.env   = None # needed for the variant when searching for files
 
-		# include paths
-		if strpaths is None:
-			self.strpaths = []
-		else:
-			self.strpaths = strpaths
-		self.pathcontents = {}
-
 		self.count_files = 0
-		self.deps  = []
-		self.deps_paths = []
+		self.currentnode_stack = []
 
-		self.m_nodepaths = nodepaths or []
-		#self.m_nodepaths.append(Params.g_build.m_root.find_dir('/usr/include'))
+		self.nodepaths = nodepaths or []
+		#self.nodepaths.append(Build.bld.root.find_dir('/usr/include'))
 
-		self.m_nodes = []
-		self.m_names = []
+		self.nodes = []
+		self.names = []
 
 		# file added
 		self.curfile = ''
@@ -394,64 +396,40 @@ class c_parser(object):
 
 		# dynamic cache
 		try:
-			self.parse_cache = Params.g_build.parse_cache
+			self.parse_cache = Build.bld.parse_cache
 		except AttributeError:
-			Params.g_build.parse_cache = {}
-			self.parse_cache = Params.g_build.parse_cache
+			Build.bld.parse_cache = {}
+			self.parse_cache = Build.bld.parse_cache
 
 	def tryfind(self, filename):
 		self.curfile = filename
-		global g_findall
-		if self.m_nodepaths:
-			found = 0
-			for n in self.m_nodepaths:
-				found = n.find_resource(filename)
-				if found:
-					break
-			# second pass for unreachable folders
-			if not found and g_findall:
-				lst = filename.split('/')
-				if len(lst)>1:
-					lst=lst[:-1] # take the folders only
-					try: cache = Params.g_build.preproc_cache
-					except AttributeError: Params.g_build.preproc_cache = cache = {}
-					key = hash( (str(self.m_nodepaths), str(lst)) )
-					if not cache.get(key, None):
-						cache[key] = 1
-						for n in self.m_nodepaths:
-							node = n.find_resource(filename)
-							if node:
-								found = node
-								break
+
+		# for msvc it should be a for loop on the whole stack
+		found = self.currentnode_stack[-1].find_resource(filename)
+
+		for n in self.nodepaths:
 			if found:
-				self.m_nodes.append(found)
-				# Qt
-				if filename[-4:] != '.moc': self.addlines(found.abspath(self.env))
-			if not found:
-				if not filename in self.m_names:
-					self.m_names.append(filename)
+				break
+			found = n.find_resource(filename)
+
+		if not found:
+			if not filename in self.names:
+				self.names.append(filename)
 		else:
-			found = 0
-			for p in self.strpaths:
-				if not p in self.pathcontents.keys():
-					self.pathcontents[p] = os.listdir(p)
-				if filename in self.pathcontents[p]:
-					#print "file %s found in path %s" % (filename, p)
-					np = os.path.join(p, filename)
-					# screw Qt two times
-					if filename[-4:] != '.moc': self.addlines(np)
-					self.deps_paths.append(np)
-					found = 1
-			if not found:
-				pass
-				#error("could not find %s " % filename)
+			self.nodes.append(found)
+			if filename[-4:] != '.moc':
+				self.addlines(found)
+		return found
+
+	def addlines(self, node):
+
+		self.currentnode_stack.append(node.parent)
+		filepath = node.abspath(self.env)
 
-	def addlines(self, filepath):
 		self.count_files += 1
-		if self.count_files > 30000: raise PreprocError, "recursion limit exceeded, bailing out"
+		if self.count_files > 30000: raise PreprocError("recursion limit exceeded, bailing out")
 		pc = self.parse_cache
-		debug("reading file %r" % filepath, 'preproc')
-
+		debug('preproc: reading file %r' % filepath)
 		try:
 			lns = pc[filepath]
 		except KeyError:
@@ -462,51 +440,41 @@ class c_parser(object):
 
 		try:
 			lines = filter_comments(filepath)
-			pc[filepath] = lines # memorize the lines filtered
+			lines.append((POPFILE, ''))
+			pc[filepath] = lines # cache the lines filtered
 			self.lines = lines + self.lines
 		except IOError:
-			raise PreprocError, "could not read the file %s" % filepath
+			raise PreprocError("could not read the file %s" % filepath)
 		except Exception:
-			if Params.g_verbose > 0:
-				warning("parsing %s failed" % filepath)
+			if Logs.verbose > 0:
+				error("parsing %s failed" % filepath)
 				traceback.print_exc()
 
 	def start(self, node, env):
-		debug("scanning %s (in %s)" % (node.m_name, node.m_parent.m_name), 'preproc')
+		debug('preproc: scanning %s (in %s)' % (node.name, node.parent.name))
 
 		self.env = env
 		variant = node.variant(env)
 
-		self.addlines(node.abspath(env))
+		self.addlines(node)
 		if env['DEFLINES']:
 			self.lines = [('define', x) for x in env['DEFLINES']] + self.lines
 
 		while self.lines:
-			(type, line) = self.lines.pop(0)
+			(kind, line) = self.lines.pop(0)
+			if kind == POPFILE:
+				self.currentnode_stack.pop()
+				continue
 			try:
-				self.process_line(type, line)
-			except Exception, ex:
-				if Params.g_verbose:
-					warning("line parsing failed (%s): %s" % (str(ex), line))
+				self.process_line(kind, line)
+			except Exception, e:
+				if Logs.verbose:
+					error("line parsing failed (%s): %s" % (str(e), line))
 					traceback.print_exc()
 
-	# debug only
-	def start_local(self, filename):
-		self.addlines(filename)
-		#print self.lines
-		while self.lines:
-			(type, line) = self.lines.pop(0)
-			try:
-				self.process_line(type, line)
-			except Exception, ex:
-				if Params.g_verbose:
-					warning("line parsing failed (%s): %s" % (str(ex), line))
-					traceback.print_exc()
-				raise
-
 	def process_line(self, token, line):
-		ve = Params.g_verbose
-		if ve: debug("line is %s - %s state is %s" % (token, line, self.state), 'preproc')
+		ve = Logs.verbose
+		if ve: debug('preproc: line is %s - %s state is %s' % (token, line, self.state))
 		state = self.state
 
 		# make certain we define the state if we are about to enter in an if block
@@ -533,13 +501,11 @@ class c_parser(object):
 			if m and m.group(0) in self.defs: state[-1] = ignored
 			else: state[-1] = accepted
 		elif token == 'include' or token == 'import':
-			(type, inc) = extract_include(line, self.defs)
+			(kind, inc) = extract_include(line, self.defs)
 			if inc in self.ban_includes: return
 			if token == 'import': self.ban_includes.append(inc)
-			if ve: debug("include found %s    (%s) " % (inc, type), 'preproc')
-			if type == '"' or not strict_quotes:
-				if not inc in self.deps:
-					self.deps.append(inc)
+			if ve: debug('preproc: include found %s    (%s) ' % (inc, kind))
+			if kind == '"' or not strict_quotes:
 				self.tryfind(inc)
 		elif token == 'elif':
 			if state[-1] == accepted:
@@ -554,10 +520,10 @@ class c_parser(object):
 			m = re_mac.search(line)
 			if m:
 				name = m.group(0)
-				if ve: debug("define %s   %s" % (name, line), 'preproc')
+				if ve: debug('preproc: define %s   %s' % (name, line))
 				self.defs[name] = line
 			else:
-				raise PreprocError, "invalid define line %s" % line
+				raise PreprocError("invalid define line %s" % line)
 		elif token == 'undef':
 			m = re_mac.search(line)
 			if m and m.group(0) in self.defs:
@@ -573,34 +539,48 @@ def extract_macro(txt):
 		p, name = t[0]
 
 		p, v = t[1]
-		if p != OP: raise PreprocError, "expected open parenthesis"
+		if p != OP: raise PreprocError("expected open parenthesis")
 
 		i = 1
 		pindex = 0
 		params = {}
-		wantident = 1
+		prev = '('
 
 		while 1:
 			i += 1
 			p, v = t[i]
 
-			if wantident:
+			if prev == '(':
 				if p == IDENT:
 					params[v] = pindex
 					pindex += 1
-				elif v == '...':
-					pass
-				else:
-					raise PreprocError, "expected ident"
-			else:
-				if v == ',':
-					pass
-				elif v == ')':
+					prev = p
+				elif p == OP and v == ')':
 					break
-				elif v == '...':
-					raise PreprocError, "not implemented"
-			wantident = not wantident
+				else:
+					raise PreprocError("unexpected token")
+			elif prev == IDENT:
+				if p == OP and v == ',':
+					prev = v
+				elif p == OP and v == ')':
+					break
+				else:
+					raise PreprocError("comma or ... expected")
+			elif prev == ',':
+				if p == IDENT:
+					params[v] = pindex
+					pindex += 1
+					prev = p
+				elif p == OP and v == '...':
+					raise PreprocError("not implemented")
+				else:
+					raise PreprocError("comma or ... expected")
+			elif prev == '...':
+				raise PreprocError("not implemented")
+			else:
+				raise PreprocError("unexpected else")
 
+		#~ print (name, [params, t[i+1:]])
 		return (name, [params, t[i+1:]])
 	else:
 		(p, v) = t[0]
@@ -614,13 +594,14 @@ def extract_include(txt, defs):
 		if m.group('b'): return '"', m.group('b')
 
 	# perform preprocessing and look at the result, it must match an include
+	tokens = tokenize(txt)
 	tokens = process_tokens(tokens, defs, ['waf_include'])
 	p, v = tokens[0]
-	if p != STR: raise PreprocError, "could not parse include %s" % txt
+	if p != STR: raise PreprocError("could not parse include %s" % txt)
 	return ('"', v)
 
 def parse_char(txt):
-	if not txt: raise PreprocError, "attempted to parse a null char"
+	if not txt: raise PreprocError("attempted to parse a null char")
 	if txt[0] != '\\':
 		return ord(txt)
 	c = txt[1]
@@ -634,7 +615,7 @@ def parse_char(txt):
 				return (1+i, int(txt[1:1+i], 8))
 	else:
 		try: return chr_esc[c]
-		except KeyError: raise PreprocError, "could not parse char literal '%s'" % txt
+		except KeyError: raise PreprocError("could not parse char literal '%s'" % txt)
 
 def tokenize(s):
 	ret = []
@@ -671,12 +652,44 @@ def tokenize(s):
 
 # quick test #
 if __name__ == "__main__":
-	Params.g_verbose = 2
-	Params.g_zones = ['preproc']
+	# first we need to replace a method for the command-line test - no nodes
+	def tryfind(self):
+		self.curfile = filename
+		found = 0
+		for p in self.strpaths:
+			if not p in self.pathcontents.keys():
+				self.pathcontents[p] = os.listdir(p)
+			if filename in self.pathcontents[p]:
+				#print "file %s found in path %s" % (filename, p)
+				np = os.path.join(p, filename)
+				# screw Qt two times
+				if filename[-4:] != '.moc': self.addlines(np)
+				found = 1
+	c_parser.tryfind = tryfind
+
+	def start_local(self, filename):
+		self.addlines(filename)
+		#print self.lines
+		while self.lines:
+			(kind, line) = self.lines.pop(0)
+			if kind == POPFILE:
+				self.currentnode_stack.pop()
+				continue
+			try:
+				self.process_line(kind, line)
+			except Exception, e:
+				if Logs.verbose:
+					error("line parsing failed (%s): %s" % (str(e), line))
+					traceback.print_exc()
+				raise e
+	c_parser.start_local = start_local
+
+	Logs.verbose = 2
+	Logs.zones = ['preproc']
 	class dum:
 		def __init__(self):
 			self.parse_cache = {}
-	Params.g_build = dum()
+	Build.bld = dum()
 
 	try: arg = sys.argv[1]
 	except IndexError: arg = "file.c"
@@ -707,11 +720,14 @@ if __name__ == "__main__":
 	test("fun2(2, fun1(m, 1))")
 	#test("foo##.##h")
 
-	gruik = c_parser(strpaths = paths)
+	gruik = c_parser()
+	gruik.strpaths = paths
+	gruik.pathcontents = {}
+
 	gruik.start_local(arg)
-	print "we have found the following dependencies"
-	print gruik.deps
-	print gruik.deps_paths
+	print("we have found the following dependencies")
+	print(gruik.nodes)
+	print(gruik.names)
 
 	#f = open(arg, "r")
 	#txt = f.read()
